@@ -1,17 +1,17 @@
 # -*- coding: utf-8 -*-
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
-import os
-import sys
-import unittest
 import doctest
-import re
-import subprocess
-import time
-from itertools import chain
-import operator
-from functools import wraps
 import inspect
+import operator
+import os
+import subprocess
+import sys
+import time
+import unittest
+from functools import reduce
+from functools import wraps
+from itertools import chain
 try:
     import pkg_resources
 except ImportError:
@@ -22,7 +22,7 @@ from sql import Table
 
 from trytond.pool import Pool, isregisteredby
 from trytond import backend
-from trytond.model import Workflow, ModelSQL, ModelSingleton, fields
+from trytond.model import Workflow, ModelSQL, ModelSingleton, ModelView, fields
 from trytond.model.fields import get_eval_fields, Function
 from trytond.tools import is_instance_method
 from trytond.transaction import Transaction
@@ -108,15 +108,10 @@ def backup_db_cache(name):
 
 
 def _db_cache_file(path, name, backend_name):
-    return os.path.join(path, '%s-%s-py%s.dump'
-        % (name, backend_name, sys.version_info.major))
+    return os.path.join(path, '%s-%s.dump' % (name, backend_name))
 
 
 def _sqlite_copy(file_, restore=False):
-    try:
-        import sqlitebck
-    except ImportError:
-        return False
     import sqlite3 as sqlite
 
     with Transaction().start(DB_NAME, 0, _nocache=True) as transaction, \
@@ -127,7 +122,14 @@ def _sqlite_copy(file_, restore=False):
             return False
         if restore:
             conn2, conn1 = conn1, conn2
-        sqlitebck.copy(conn1, conn2)
+        if hasattr(conn1, 'backup'):
+            conn1.backup(conn2)
+        else:
+            try:
+                import sqlitebck
+            except ImportError:
+                return False
+            sqlitebck.copy(conn1, conn2)
     return True
 
 
@@ -289,7 +291,7 @@ class ModuleTestCase(unittest.TestCase):
         for mname, model in Pool().iterobject():
             if not isregisteredby(model, self.module):
                 continue
-            for fname, field in model._fields.iteritems():
+            for fname, field in model._fields.items():
                 fields = set()
                 fields |= get_eval_fields(field.domain)
                 if hasattr(field, 'digits'):
@@ -308,6 +310,33 @@ class ModuleTestCase(unittest.TestCase):
                 assert depends <= set(model._fields), (
                     'Unknown depends %s in "%s"."%s"' % (
                         list(depends - set(model._fields)), mname, fname))
+            if issubclass(model, ModelView):
+                for bname, button in model._buttons.items():
+                    depends = set(button.get('depends', []))
+                    assert depends <= set(model._fields), (
+                        'Unknown depends %s in button "%s"."%s"' % (
+                            list(depends - set(model._fields)), mname, bname))
+
+    @with_transaction()
+    def test_depends_parent(self):
+        "Test depends on _parent_ contains also the parent relation"
+        for mname, model in Pool().iterobject():
+            if not isregisteredby(model, self.module):
+                continue
+            for fname, field in model._fields.items():
+                for attribute in ['depends', 'on_change', 'on_change_with',
+                        'selection_change_with', 'autocomplete']:
+                    depends = getattr(field, attribute, [])
+                    for depend in depends:
+                        prefix = []
+                        for d in depend.split('.'):
+                            if d.startswith('_parent_'):
+                                relation = '.'.join(
+                                    prefix + [d[len('_parent_'):]])
+                                assert relation in depends, (
+                                    'Missing "%s" in "%s"."%s"."%s"' % (
+                                        relation, mname, fname, attribute))
+                            prefix.append(d)
 
     @with_transaction()
     def test_field_methods(self):
@@ -443,7 +472,7 @@ class ModuleTestCase(unittest.TestCase):
                     'wizard': wizard_name,
                     })
             wizard_instance = wizard(session_id)
-            for state_name, state in wizard_instance.states.iteritems():
+            for state_name, state in wizard_instance.states.items():
                 if isinstance(state, StateView):
                     # Don't test defaults as they may depend on context
                     state.get_view(wizard_instance, state_name)
@@ -457,7 +486,7 @@ class ModuleTestCase(unittest.TestCase):
         for mname, model in Pool().iterobject():
             if not isregisteredby(model, self.module):
                 continue
-            for field_name, field in model._fields.iteritems():
+            for field_name, field in model._fields.items():
                 selection = getattr(field, 'selection', None)
                 if selection is None:
                     continue
@@ -483,7 +512,7 @@ class ModuleTestCase(unittest.TestCase):
         for mname, model in Pool().iterobject():
             if not isregisteredby(model, self.module):
                 continue
-            for field_name, field in model._fields.iteritems():
+            for field_name, field in model._fields.items():
                 if not isinstance(field, Function):
                     continue
                 for func_name in [field.getter, field.setter, field.searcher]:
@@ -511,20 +540,34 @@ class ModuleTestCase(unittest.TestCase):
             if not action_window.res_model:
                 continue
             Model = pool.get(action_window.res_model)
-            decoder = PYSONDecoder({
-                    'active_id': None,
-                    'active_ids': [],
-                    'active_model': action_window.res_model,
-                    })
-            domain = decoder.decode(action_window.pyson_domain)
-            order = decoder.decode(action_window.pyson_order)
-            context = decoder.decode(action_window.pyson_context)
-            with Transaction().set_context(context):
-                Model.search(domain, order=order, limit=action_window.limit)
-            for action_domain in action_window.act_window_domains:
-                if not action_domain.domain:
-                    continue
-                Model.search(decoder.decode(action_domain.domain))
+            for active_id, active_ids in [
+                    (None, []),
+                    (1, [1]),
+                    (1, [1, 2]),
+                    ]:
+                decoder = PYSONDecoder({
+                        'active_id': active_id,
+                        'active_ids': active_ids,
+                        'active_model': action_window.res_model,
+                        })
+                domain = decoder.decode(action_window.pyson_domain)
+                order = decoder.decode(action_window.pyson_order)
+                context = decoder.decode(action_window.pyson_context)
+                search_value = decoder.decode(action_window.pyson_search_value)
+                if action_window.context_domain:
+                    domain = ['AND', domain,
+                        decoder.decode(action_window.context_domain)]
+                with Transaction().set_context(context):
+                    Model.search(
+                        domain, order=order, limit=action_window.limit)
+                    if search_value:
+                        Model.search(search_value)
+                for action_domain in action_window.act_window_domains:
+                    if not action_domain.domain:
+                        continue
+                    Model.search(decoder.decode(action_domain.domain))
+            if action_window.context_model:
+                pool.get(action_window.context_model)
 
     @with_transaction()
     def test_modelsingleton_inherit_order(self):
@@ -542,6 +585,27 @@ class ModuleTestCase(unittest.TestCase):
                 "ModelSingleton must appear before ModelSQL in the parent "
                 "classes of '%s'." % mname)
 
+    @with_transaction()
+    def test_buttons_registered(self):
+        'Test all buttons are registered in ir.model.button'
+        pool = Pool()
+        Button = pool.get('ir.model.button')
+        for mname, model in Pool().iterobject():
+            if not isregisteredby(model, self.module):
+                continue
+            if not issubclass(model, ModelView):
+                continue
+            ir_buttons = {b.name for b in Button.search([
+                        ('model.model', '=', model.__name__),
+                        ])}
+            buttons = set(model._buttons)
+            assert ir_buttons >= buttons, (
+                'The buttons "%(buttons)s" of Model "%(model)s" are not '
+                'registered in ir.model.button.' % {
+                    'buttons': list(buttons - ir_buttons),
+                    'model': model.__name__,
+                    })
+
 
 def db_exist(name=DB_NAME):
     Database = backend.get('Database')
@@ -599,18 +663,16 @@ def drop_create(name=DB_NAME, lang='en'):
         drop_db(name)
     create_db(name, lang)
 
-doctest_setup = lambda test: drop_create()
-doctest_teardown = lambda test: drop_db()
+
+def doctest_setup(test):
+    return drop_create()
 
 
-class Py23DocChecker(doctest.OutputChecker):
-    def check_output(self, want, got, optionflags):
-        if sys.version_info[0] > 2:
-            want = re.sub("u'(.*?)'", "'\\1'", want)
-            want = re.sub('u"(.*?)"', '"\\1"', want)
-        return doctest.OutputChecker.check_output(self, want, got, optionflags)
+def doctest_teardown(test):
+    return drop_db()
 
-doctest_checker = Py23DocChecker()
+
+doctest_checker = doctest.OutputChecker()
 
 
 class TestSuite(unittest.TestSuite):
@@ -620,7 +682,7 @@ class TestSuite(unittest.TestSuite):
             try:
                 exist = db_exist()
                 break
-            except DatabaseOperationalError, err:
+            except DatabaseOperationalError as err:
                 # Retry on connection error
                 sys.stderr.write(str(err))
                 time.sleep(1)
@@ -675,21 +737,16 @@ def modules_suite(modules=None, doc=True):
         suite_ = suite()
     else:
         suite_ = all_suite()
-    from trytond.modules import create_graph, get_module_list, \
-        MODULES_PATH, EGG_MODULES
-    graph = create_graph(get_module_list())[0]
-    for package in graph:
-        module = package.name
+    from trytond.modules import create_graph, get_module_list, import_module
+    graph = create_graph(get_module_list())
+    for node in graph:
+        module = node.name
         if modules and module not in modules:
             continue
         test_module = 'trytond.modules.%s.tests' % module
-        if os.path.isdir(os.path.join(MODULES_PATH, module)) or \
-                module in EGG_MODULES:
-            try:
-                test_mod = __import__(test_module, fromlist=[''])
-            except ImportError:
-                continue
-        else:
+        try:
+            test_mod = import_module(module, test_module)
+        except ImportError:
             continue
         for test in test_mod.suite():
             if isinstance(test, doctest.DocTestCase) and not doc:
